#include "asvariant.h"
#include <string.h>
#include <AS3.h>

/**
 * RXgN^
 */
AS3Variant::AS3Variant() : obj(NULL), str(NULL)
{
}

AS3Variant::AS3Variant(bool n) : str(NULL)
{
	obj = n ? AS3_True() : AS3_False();
}

AS3Variant::AS3Variant(int n) : str(NULL)
{
	obj = AS3_Int(n);
}

AS3Variant::AS3Variant(float n) : str(NULL)
{
	obj = AS3_Number(n);
}

AS3Variant::AS3Variant(double n) : str(NULL)
{
	obj = AS3_Number(n);
}

AS3Variant::AS3Variant(const char *str) : str(NULL)
{
	obj = AS3_String(str);
}

AS3Variant::AS3Variant(AS3_Val val) : obj(val), str(NULL)
{
	obj = val;
}

AS3Variant::AS3Variant(const AS3Variant &orig) : obj(orig.obj), str(NULL)
{
	AS3_Acquire(obj);
}

/**
 * ej undefined ɂ
 */
void
AS3Variant::clear()
{
	AS3_Release(obj);
	obj = NULL;
	free(str);
	str = NULL;
}
	
/**
 * fXgN^
 */
AS3Variant::~AS3Variant()
{
	clear();
}

// 
AS3Variant& AS3Variant::operator=(const AS3Variant &orig)
{
	clear();
	obj = orig.obj;
	AS3_Acquire(obj);
	return *this;
}

// r
bool AS3Variant::operator==(const AS3Variant &target)
{
	AS3Variant equals = createFunction(APPNAMESPACE "equals");
	return (int)equals.call(*this, target);
}

AS3Variant::operator int() const
{
	return AS3_IntValue(obj);
}

AS3Variant::operator double() const
{
	return AS3_NumberValue(obj);
}

AS3Variant::operator float() const
{
	return (float)AS3_NumberValue(obj);
}

AS3Variant::operator const char*()
{
	if (str == NULL) {
		str = AS3_StringValue(obj);
	}
	return str;
}

AS3Variant::operator char*()
{
	return AS3_StringValue(obj);
}

/**
 * ByteArray ̃oCgǂݏo
 * @param buffer i[obt@
 * @param size i[TCY
 */
void
AS3Variant::readBytes(char *buffer, int size)
{
	AS3_ByteArray_readBytes(buffer, obj, size);
}

// g[X\
void
AS3Variant::trace()
{
	AS3_Trace(obj);
}

/**
 * @return ^
 */
AS3Variant::Type
AS3Variant::getType() const
{
	AS3Variant getType = createFunction(APPNAMESPACE "getType");
	return (Type)(int)getType.call(*this);
}

/**
 * @return `Ȃ true
 */
bool
AS3Variant::isUndefined() const
{
	return obj == NULL;
}

/**
 *@return NXIuWFNgȂ true
 */
bool
AS3Variant::isClass() const
{
	AS3Variant isClass = createFunction(APPNAMESPACE "isClass");
	return (int)isClass.call(*this) != 0;
}

/**
 *@return t@NVȂ true
 */
bool
AS3Variant::isFunction() const
{
	return getType() == TYPE_FUNCTION;
}

/**
 * @reutrn IuWFNgȂ true
 */
bool
AS3Variant::isObject() const
{
	return getType() == TYPE_OBJECT;
}

/**
 * AS3NX擾
 * @param instance CX^X
 * @return NX
 */
AS3Variant
AS3Variant::getClassName() const
{
	AS3Variant func = createFunction("flash.utils.getQualifiedClassName");
	return func.call(*this);
}

// -------------------------------------------------------
// setter/getter
// -------------------------------------------------------

/**
 * l擾
 * @param obj ΏƃIuWFNg
 * @param o
 * @return o̒lintŕԂ
 */
AS3Variant
AS3Variant::get(const AS3Variant &name) const
{
	return AS3Variant(AS3_Get(obj, name.obj));
}

/**
 * lݒ
 * @param obj ΏƃIuWFNg
 * @param o
 * @param value ݒl
 */
void
AS3Variant::set(const AS3Variant &name, const AS3Variant &value)
{
	AS3_Release(AS3_Set(obj, name.obj, value.obj));
}

/**
 * zƂ݂Ȃăp[^ǉ
 */
void
AS3Variant::arrayPush(const AS3Variant &value)
{
	AS3_Release(AS3_CallTS("push", obj, "AS3ValType", value.obj));
}

// --------------------------------------------------------------

/**
 * t@NVĂяo
 * @param args Xg
 * @return Al
 */
AS3Variant
AS3Variant::call(AS3Variant *args)
{
	if (args && args->obj) {
		return AS3Variant(AS3_Call(obj, NULL, args->obj));
	} else {
		AS3Variant a = createArray();
		return AS3Variant(AS3_Call(obj, NULL, a.obj));
	}
}

/**
 * \bhĂяo(\bhw)
 * @param methodName \bh
 * @param args 
 */
AS3Variant
AS3Variant::callMethod(const char *methodName, AS3Variant *args)
{
	if (args && args->obj) {
		return AS3Variant(AS3_CallS(methodName, obj, args->obj));
	} else {
		AS3Variant a = createArray();
		return AS3Variant(AS3_CallS(methodName, obj, a.obj));
	}
}

/**
 * \bhĂяo(\bhڎw)
 * @param method \bhIuWFNg
 * @param args 
 */
AS3Variant
AS3Variant::callMethod(const AS3Variant &method, AS3Variant *args)
{
	if (args && args->obj) {
		return AS3Variant(AS3_Call(method.obj, obj, args != NULL ? args->obj : NULL));
	} else {
		AS3Variant a = createArray();
		return AS3Variant(AS3_Call(method.obj, obj, a.obj));
	}
}

/**
 * NX̃CX^X
 * @param args 
 * @return ꂽCX^X
 */
AS3Variant
AS3Variant::createInstance(AS3Variant *args)
{
	if (args && args->obj) {
		return AS3Variant(AS3_New(obj, args->obj));
	} else {
		AS3Variant a = createArray();
		return AS3Variant(AS3_New(obj, a.obj));
	}
}

// -------------------------------------------------------
// ꏉ
// -------------------------------------------------------

/**
 * O[o֐ŏ
 */
AS3Variant
AS3Variant::createFunction(const char *funcName)
{
	AS3_Val ns;
	const char *func = strrchr(funcName, '.');
	if (func) {
		ns = AS3_StringN(funcName, func - funcName);
		funcName = func + 1;
	} else {
		ns = AS3_String("");
	}
	AS3Variant ret(AS3_NSGetS(ns, funcName));
	AS3_Release(ns);
	return ret;
}

/**
 * AS3
 * @param className NX
 */
AS3Variant
AS3Variant::createDictionary()
{
	return AS3Variant(AS3_Object(""));
}

/**
 * AS3z񐶐
 * @param className NX
 */

AS3Variant
AS3Variant::createArray()
{
	return AS3Variant(AS3_Array(""));
}

AS3_Val
AS3Variant::getObj() {
	AS3_Acquire(obj);
	return obj;
}
