#include <new>
#include <stdio.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>

#include "sqwrapper.h"
#include "sqas3obj.h"

/**
 * Flashg[XOo͗p
 */
void trace(const char *format, ...)
{
	va_list args;
	va_start(args, format);
	char msg[1024];
	vsnprintf(msg, sizeof msg-1, format, args);
	AS3Variant m(msg);
	m.trace();
	va_end(args);
}

// ActionScript Object ʗp
static const SQUserPointer AS3_OBJ_TAG = (SQUserPointer)"AS3_OBJ_TAG";

/**
 * IuWFNg
 */
static SQInteger
objRelease(SQUserPointer up, SQInteger size)
{
	AS3Variant *self = (AS3Variant*)up;
	if (self) {
		self->~AS3Variant();
		sq_free(up, size);
	}
	return 1;
}

/**
 * ActionScript3 p getter
 * @param v squirrel VM
 */
static SQInteger
get(HSQUIRRELVM v)
{
	AS3Variant self = createVariant(v,1);
	AS3Variant ret = self.get(sqobject::getString(v,2));
	return ret.isUndefined() ? SQ_ERROR : sq_pushvariant(v, ret);
}

/**
 * ActionScript3 p setter
 * @param v squirrel VM
 */
static SQInteger
set(HSQUIRRELVM v)
{
	AS3Variant self = createVariant(v,1);
	self.set(sqobject::getString(v,2), createVariant(v,3));
	return SQ_OK;
}

/**
 * ActionScript3 p\bh܂̓RXgN^̌Ăяo
 * @param v squirrel VM
 */
static SQInteger
call(HSQUIRRELVM v)
{
	// param1 callΏ
	// param2 this
	// param3 ` {̈

	//trace("start call");
	AS3Variant func = createVariant(v,1);
	AS3Variant args = createArray(v, 3, sq_gettop(v));
	if (func.getType() == AS3Variant::TYPE_FUNCTION) {
		AS3Variant self = createVariant(v,2);
		AS3Variant ret = self.callMethod(func, &args);
		return sq_pushvariant(v, ret);
	} else if (func.isClass())  {
		AS3Variant ret = func.createInstance(&args);
		return sq_pushvariant(v, ret);
	}
	// ĂяołȂ
	return sq_throwerror(v, "invalid call");
}

/**
 * ActionScript3 IuWFNgX^bNɊi[
 * @param v squirrel VM
 * @param obj ActionScript3 IuWFNg
 */
SQRESULT
sq_pushvariant(HSQUIRRELVM v, AS3Variant &variant)
{
	if (variant.isUndefined()) {
		return 0;
	}
	switch (variant.getType()) {
	case AS3Variant::TYPE_NULL:
		sq_pushnull(v);
		break;
	case AS3Variant::TYPE_BOOLEAN:
		sq_pushbool(v, (int)variant ? SQTrue : SQFalse);
		break;
	case AS3Variant::TYPE_INTEGER:
		sq_pushinteger(v, (SQInteger)variant);
		break;
	case AS3Variant::TYPE_NUMBER:
		sq_pushfloat(v, (SQFloat)variant);
		break;
	case AS3Variant::TYPE_STRING:
		sq_pushstring(v, (const char*)variant, -1);
		break;
	default:
		//trace("push object:");
		if (!AS3Object::pushInstance(v, variant)) {
			
			AS3Variant *self = (AS3Variant*)sq_newuserdata(v, sizeof *self);
			new (self) AS3Variant(variant);
			
			// ^Oo^
			sq_settypetag(v, -1, AS3_OBJ_TAG);
			
			// JWbNǉ
			sq_setreleasehook(v, -1, objRelease);
			
			// \bhQǉ
			sq_newtable(v);
		
			sq_pushstring(v, "_get", -1);
			sq_newclosure(v, get, 0);
			sq_createslot(v, -3);
			
			sq_pushstring(v, "_set", -1);
			sq_newclosure(v, set, 0);
			sq_createslot(v, -3);

			sq_pushstring(v, "_call", -1);
			sq_newclosure(v, call, 0);
			sq_createslot(v, -3);
			
			sq_setdelegate(v, -2);
		}
	}
	return 1;
}


/**
 * squirrelX^bNAS3Variant擾
 * @param v VM
 * @param idx X^bÑCfbNX
 * @param result IuWFNgԂ
 */
AS3Variant
createVariant(HSQUIRRELVM v, SQInteger idx)
{
	switch (sq_gettype(v,idx)) {
	case OT_INTEGER:
		{
			SQInteger val;
			sq_getinteger(v, idx, &val);
			return AS3Variant(val);
		}
		break;
	case OT_FLOAT:
		{
			SQFloat val;
			sq_getfloat(v, idx, &val);
			return AS3Variant(val);
		}
		break;
	case OT_BOOL:
		{
			SQBool val;
			sq_getbool(v, idx, &val);
			return AS3Variant(val != SQFalse);
		}
		break;
	case OT_STRING:
		{
			const SQChar *val;
			sq_getstring(v, idx, &val);
			return AS3Variant(val);
		}
		break;
	case OT_USERDATA:
		{
			SQUserPointer otag;
			SQUserPointer up;
			//trace("from userdata");
			if (SQ_SUCCEEDED(sq_getuserdata(v,idx,&up,&otag)) && otag == AS3_OBJ_TAG) {
				return *((AS3Variant*)up);
			} else {
				return AS3Variant();
			}
		}
		break;
	case OT_CLASS:
	case OT_INSTANCE:
		{
			AS3Variant ret;
			AS3Object::getInstance(v, idx, ret);
			return ret;
		}
	default:
		break;
	}
	
	return AS3Variant();
}

/**
 * AS3pXg̐
 * @param v squirrelvm
 * @param idx JnCfbNX
 * @param max ICfbNX
 * @return AS3Xg̎
 */
AS3Variant
createArray(HSQUIRRELVM v, SQInteger idx, SQInteger max)
{
	AS3Variant ret = AS3Variant::createArray();
	for (SQInteger i=idx;i<=max;i++) {
		ret.arrayPush(createVariant(v,i));
	}
	return ret;
}
